#!/usr/bin/env python

"""
Interpretation and preparation of iMIP content, together with a content handling
mechanism employed by specific recipients.

Copyright (C) 2014, 2015 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from datetime import datetime, timedelta
from email.mime.text import MIMEText
from imiptools.config import MANAGER_PATH, MANAGER_URL
from imiptools.data import Object, parse_object, \
                           get_address, get_uri, get_value, \
                           is_new_object, uri_dict, uri_item
from imiptools.dates import *
from imiptools.period import can_schedule, insert_period, remove_period
from pytz import timezone
from socket import gethostname
from vRecurrence import get_parameters, get_rule
import imip_store

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

# NOTE: Need to expose the 100 day window for recurring events in the
# NOTE: configuration.

def get_periods(obj, window_size=100):

    """
    Return periods for the given object 'obj', confining materialised periods
    to the given 'window_size' in days starting from the present moment.
    """

    dtstart = obj.get_utc_datetime("DTSTART")
    dtend = obj.get_utc_datetime("DTEND")

    # NOTE: Need also DURATION support.

    duration = dtend - dtstart

    # Recurrence rules create multiple instances to be checked.
    # Conflicts may only be assessed within a period defined by policy
    # for the agent, with instances outside that period being considered
    # unchecked.

    window_end = datetime.now() + timedelta(window_size)

    # NOTE: Need also RDATE and EXDATE support.

    rrule = obj.get_value("RRULE")

    if rrule:
        selector = get_rule(dtstart, rrule)
        parameters = get_parameters(rrule)
        periods = []
        for start in selector.materialise(dtstart, window_end, parameters.get("COUNT"), parameters.get("BYSETPOS")):
            start = datetime(*start, tzinfo=timezone("UTC"))
            end = start + duration
            periods.append((format_datetime(start), format_datetime(end)))
    else:
        periods = [(format_datetime(dtstart), format_datetime(dtend))]

    return periods

def remove_from_freebusy(freebusy, attendee, uid, store):

    """
    For the given 'attendee', remove periods from 'freebusy' that are associated
    with 'uid' in the 'store'.
    """

    remove_period(freebusy, uid)
    store.set_freebusy(attendee, freebusy)

def remove_from_freebusy_for_other(freebusy, user, other, uid, store):

    """
    For the given 'user', remove for the 'other' party periods from 'freebusy'
    that are associated with 'uid' in the 'store'.
    """

    remove_period(freebusy, uid)
    store.set_freebusy_for_other(user, freebusy, other)

def _update_freebusy(freebusy, periods, transp, uid):

    """
    Update the free/busy details with the given 'periods', 'transp' setting and
    'uid'.
    """

    remove_period(freebusy, uid)

    for start, end in periods:
        insert_period(freebusy, (start, end, uid, transp))

def update_freebusy(freebusy, attendee, periods, transp, uid, store):

    """
    For the given 'attendee', update the free/busy details with the given
    'periods', 'transp' setting and 'uid' in the 'store'.
    """

    _update_freebusy(freebusy, periods, transp, uid)
    store.set_freebusy(attendee, freebusy)

def update_freebusy_for_other(freebusy, user, other, periods, transp, uid, store):

    """
    For the given 'user', update the free/busy details of 'other' with the given
    'periods', 'transp' setting and 'uid' in the 'store'.
    """

    _update_freebusy(freebusy, periods, transp, uid)
    store.set_freebusy_for_other(user, freebusy, other)

# Handler mechanism objects.

def handle_itip_part(part, handlers):

    """
    Handle the given iTIP 'part' using the given 'handlers'.

    Return a list of responses, each response being a tuple of the form
    (outgoing-recipients, message-part).
    """

    handlers = dict(handlers)
    method = part.get_param("method")

    # Decode the data and parse it.

    f = StringIO(part.get_payload(decode=True))

    itip = parse_object(f, part.get_content_charset(), "VCALENDAR")

    # Ignore the part if not a calendar object.

    if not itip:
        return []

    # Require consistency between declared and employed methods.

    if get_value(itip, "METHOD") == method:

        # Look for different kinds of sections.

        all_results = []

        for name, items in itip.items():

            # Get a handler for the given section.

            handler = handlers.get(name)
            if not handler:
                continue

            for item in items:

                # Dispatch to a handler and obtain any response.

                handler.set_object(Object({name : item}))
                results = methods[method](handler)()

                # Aggregate responses for a single message.

                if results:
                    for result in results:
                        outgoing_recipients, part = result
                        all_results.append((outgoing_recipients, part))

        return all_results

    return []

# References to the Web interface.

def get_manager_url():
    url_base = MANAGER_URL or "http://%s/" % gethostname()
    return "%s/%s" % (url_base.rstrip("/"), MANAGER_PATH.lstrip("/"))

def get_object_url(uid):
    return "%s/%s" % (get_manager_url().rstrip("/"), uid)

class Handler:

    "General handler support."

    def __init__(self, senders=None, recipient=None, messenger=None):

        """
        Initialise the handler with the calendar 'obj' and the 'senders' and
        'recipient' of the object (if specifically indicated).
        """

        self.senders = senders and set(map(get_address, senders))
        self.recipient = recipient and get_address(recipient)
        self.messenger = messenger

        self.obj = None
        self.uid = None
        self.sequence = None
        self.dtstamp = None

        self.store = imip_store.FileStore()

        try:
            self.publisher = imip_store.FilePublisher()
        except OSError:
            self.publisher = None

    def set_object(self, obj):
        self.obj = obj
        self.uid = self.obj.get_value("UID")
        self.sequence = self.obj.get_value("SEQUENCE")
        self.dtstamp = self.obj.get_value("DTSTAMP")

    def wrap(self, text, link=True):

        "Wrap any valid message for passing to the recipient."

        texts = []
        texts.append(text)
        if link:
            texts.append("If your mail program cannot handle this "
                         "message, you may view the details here:\n\n%s" %
                         get_object_url(self.uid))

        return [(None, MIMEText("\n".join(texts)))]

    # Access to calendar structures and other data.

    def get_periods(self):
        return get_periods(self.obj)

    def remove_from_freebusy(self, freebusy, attendee):
        remove_from_freebusy(freebusy, attendee, self.uid, self.store)

    def remove_from_freebusy_for_other(self, freebusy, user, other):
        remove_from_freebusy_for_other(freebusy, user, other, self.uid, self.store)

    def update_freebusy(self, freebusy, attendee, periods):
        return update_freebusy(freebusy, attendee, periods, self.obj.get_value("TRANSP"), self.uid, self.store)

    def update_freebusy_for_other(self, freebusy, user, other, periods):
        return update_freebusy_for_other(freebusy, user, other, periods, self.obj.get_value("TRANSP"), self.uid, self.store)

    def can_schedule(self, freebusy, periods):
        return can_schedule(freebusy, periods, self.uid)

    def filter_by_senders(self, mapping):

        """
        Return a list of items from 'mapping' filtered using sender information.
        """

        if self.senders:

            # Get a mapping from senders to identities.

            identities = self.get_sender_identities(mapping)

            # Find the senders that are valid.

            senders = map(get_address, identities)
            valid = self.senders.intersection(senders)

            # Return the true identities.

            return [identities[get_uri(address)] for address in valid]
        else:
            return mapping

    def filter_by_recipient(self, mapping):

        """
        Return a list of items from 'mapping' filtered using recipient
        information.
        """

        if self.recipient:
            addresses = set(map(get_address, mapping))
            return map(get_uri, addresses.intersection([self.recipient]))
        else:
            return mapping

    def require_organiser_and_attendees(self, from_organiser=True):

        """
        Return the organiser and attendees for the current object, filtered for
        the recipient of interest. Return None if no identities are eligible.

        Organiser and attendee identities are provided as lower case values.
        """

        attendee_map = uri_dict(self.obj.get_value_map("ATTENDEE"))
        organiser_item = uri_item(self.obj.get_item("ORGANIZER"))

        # Only provide details for attendees who sent/receive the message.

        attendee_filter_fn = from_organiser and self.filter_by_recipient or self.filter_by_senders

        attendees = {}
        for attendee in attendee_filter_fn(attendee_map):
            attendees[attendee] = attendee_map[attendee]

        if not attendees or not organiser_item:
            return None

        # Only provide details for an organiser who sent/receives the message.

        organiser_filter_fn = from_organiser and self.filter_by_senders or self.filter_by_recipient

        if not organiser_filter_fn(dict([organiser_item])):
            return None

        return organiser_item, attendees

    def get_sender_identities(self, mapping):

        """
        Return a mapping from actual senders to the identities for which they
        have provided data, extracting this information from the given
        'mapping'.
        """

        senders = {}

        for value, attr in mapping.items():
            sent_by = attr.get("SENT-BY")
            if sent_by:
                senders[get_uri(sent_by)] = value
            else:
                senders[value] = value

        return senders

    def get_object(self, user):

        """
        Return the stored object to which the current object refers for the
        given 'user' and for the given 'objtype'.
        """

        f = self.store.get_event(user, self.uid)
        fragment = f and parse_object(f, "utf-8")
        return fragment and Object(fragment)

    def have_new_object(self, attendee, obj=None):

        """
        Return whether the current object is new to the 'attendee' (or if the
        given 'obj' is new).
        """

        obj = obj or self.get_object(attendee)

        # If found, compare SEQUENCE and potentially DTSTAMP.

        if obj:
            sequence = obj.get_value("SEQUENCE")
            dtstamp = obj.get_value("DTSTAMP")

            # If the request refers to an older version of the object, ignore
            # it.

            return is_new_object(sequence, self.sequence, dtstamp, self.dtstamp,
                self.is_partstat_updated(obj))

        return True

    def is_partstat_updated(self, obj):

        """
        Return whether the participant status has been updated in the current
        object in comparison to the given 'obj'.

        NOTE: Some clients like Claws Mail erase time information from DTSTAMP
        NOTE: and make it invalid. Thus, such attendance information may also be
        NOTE: incorporated into any new object assessment.
        """

        old_attendees = obj.get_value_map("ATTENDEE")
        new_attendees = self.obj.get_value_map("ATTENDEE")

        for attendee, attr in old_attendees.items():
            old_partstat = attr.get("PARTSTAT")
            new_attr = new_attendees.get(attendee)
            new_partstat = new_attr and new_attr.get("PARTSTAT")

            if old_partstat == "NEEDS-ACTION" and new_partstat and \
               new_partstat != old_partstat:

                return True

        return False

    def update_dtstamp(self):

        "Update the DTSTAMP in the current object."

        dtstamp = self.obj.get_utc_datetime("DTSTAMP")
        utcnow = to_timezone(datetime.utcnow(), "UTC")
        self.obj["DTSTAMP"] = [(format_datetime(dtstamp > utcnow and dtstamp or utcnow), {})]

# Handler registry.

methods = {
    "ADD"            : lambda handler: handler.add,
    "CANCEL"         : lambda handler: handler.cancel,
    "COUNTER"        : lambda handler: handler.counter,
    "DECLINECOUNTER" : lambda handler: handler.declinecounter,
    "PUBLISH"        : lambda handler: handler.publish,
    "REFRESH"        : lambda handler: handler.refresh,
    "REPLY"          : lambda handler: handler.reply,
    "REQUEST"        : lambda handler: handler.request,
    }

# vim: tabstop=4 expandtab shiftwidth=4
